from __future__ import annotations

import typing as typ

from dataclasses import dataclass
from numbers import Number

import numpy as np

from jijmodeling import Problem

from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    merge_params_and_kwargs,
    sample_model,
)
from jijzept.sampler.sample_qubo import sample_qubo

T = typ.TypeVar("T")


@dataclass
class JijSQAParameters:
    beta: float | None = None
    gamma: float | None = None
    trotter: int | None = None
    num_sweeps: int | None = None
    num_reads: int | None = None
    sparse: bool | None = None
    reinitialize_state: bool | None = None
    seed: int | None = None


class JijSQASampler(JijZeptBaseSampler):

    solver_type = SolverType(queue_name="openjijsolver", solver="SQA")
    jijmodeling_solver_type = SolverType(
        queue_name="openjijsolver", solver="SQAParaSearch"
    )

    def sample_qubo(
        self,
        qubo: dict,
        constant: float = 0,
        parameters: JijSQAParameters | None = None,
        timeout: int | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using BinaryQuadraticModel by means of the simulated annealing.

        Args:
            bqm (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel]): Binary quadratic model.
            parameters: (JijSAParameters | None, optional): defaults None.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
            For `cimod.BinaryQuadraticModel` case:
            ```python
            import jijzept as jz
            import cimod
            bqm = cimod.BinaryQuadraticModel({0: -1, 1: -1}, {(0, 1): -1, (1, 2): -1}, "SPIN")
            sampler = jz.JijSASampler(config='config.toml')
            response = sampler.sample(bqm)
            ```
            One can also use `sample_ising` and `sample_qubo` methods.

            For Ising case:
            ```python
            import jijzept as jz
            h = {0: -1, 1: -1, 2: 1, 3: 1}
            J = {(0, 1): -1, (3, 4): -1}
            sampler = jz.JijSASampler(config='config.toml')
            response = sampler.sample_ising(h, J)
            ```

            For QUBO case:
            ```python
            import jijzept as jz
            Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (0, 1): -1, (1, 2): 1}
            sampler = jz.JijSASampler(config='config.toml')
            response = sampler.sample_qubo(Q)
            ```
        """
        param_dict = merge_params_and_kwargs(parameters, kwargs, JijSQAParameters)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_qubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            qubo=qubo,
            constant=constant,
            timeout=timeout,
            sync=sync,
            **param_dict,
        )

        return sample_set

    def sample_model(
        self,
        problem: Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: str | None = None,
        parameters: JijSQAParameters | None = None,
        timeout: int | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of the simulated annealing.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.Binary('x', shape=n)
            d = jm.Placeholder('d', shape=n)
            i = jm.Element("i", n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijSASampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """

        param_dict = merge_params_and_kwargs(parameters, kwargs, JijSQAParameters)

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=problem,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            timeout=timeout,
            sync=sync,
            **param_dict,
        )
        return sample_set
