from __future__ import annotations

import itertools

import jijmodeling as jm
import numpy as np

from jijzept.client import JijZeptClient
from jijzept.sampler.base_sampler import ParameterSearchParameters, sample_model


def hubo_problem(
    hubo: dict[tuple[int, ...], float] | dict[tuple[str, ...], float], constant: float
) -> tuple[
    jm.Problem, dict[str, np.ndarray | list], dict[int, str], dict[str, jm.Binary]
]:

    qubo_keys = list(hubo.keys())
    unique_labels: dict[str, int] = {}
    N = jm.Placeholder("N")
    key = qubo_keys[0][0]
    if isinstance(key, int):
        # decision_variables = [jm.Binary("x", shape=N)]
        pass
    elif isinstance(key, str):
        unique_labels = {
            k: i for i, k in enumerate(itertools.chain.from_iterable(qubo_keys))
        }
        qubo_keys = [[unique_labels[e0], unique_labels[e1]] for e0, e1 in qubo_keys]
    else:
        TypeError(
            f"`{key.__class__.__name__}` object is not supported as a key of QUBO."
        )

    N = jm.Placeholder("N")  # max variable index
    E = jm.JaggedArray("E", dim=2)  # super graph edge
    L = E.shape[0]  # number of edges
    Q = jm.Placeholder("Q", shape=L)  # weight of edge
    c = jm.Placeholder("c")  # constant
    e = jm.Element("e", (0, L))  # edge index

    d = E[e].shape[0]  # order of edge 'e'
    k = jm.Element("k", (0, d))

    x = jm.Binary("x", shape=N)  # decision variable

    hubo_model = jm.Problem("qubo")
    hubo_model += jm.Sum(e, Q[e] * jm.Prod(k, x[E[e, k]])) + c

    edges = [list(k) for k in hubo.keys()]
    values = np.array(list(hubo.values()), dtype=float)
    instance_data = {
        "N": len(edges),
        "E": edges,
        "Q": values.tolist(),
        "c": constant,
    }

    return hubo_model, instance_data, {i: k for k, i in unique_labels.items()}, {"x": x}


def sample_hubo(
    client: JijZeptClient,
    solver: str,
    queue_name: str,
    polynomial: dict[tuple[int, ...], float],
    constant: float,
    timeout: float | None,
    sync: bool,
    **kwargs,
):
    hubo_model, instance_data, var_map, _ = hubo_problem(polynomial, constant)

    parameters = ParameterSearchParameters(
        multipliers={}, mul_search=False, normalize_qubo=False
    )

    sample_set = sample_model(
        client=client,
        solver=solver,
        queue_name=queue_name,
        problem=hubo_model,
        instance_data=instance_data,
        fixed_variables={},
        parameter_search_parameters=parameters,
        timeout=timeout,
        sync=sync,
        **kwargs,
    )
    sample_set.set_variable_labels(var_map)

    return sample_set
